define('framework/directive/ControlValidatorConstructor', ['angular', 'jquery'], function (angular, $) {
    return function (Arrays, $http, $timeout) {

        /**
         * 同步验证匹配器
         *
         * @param validationParams {Object} 验证必要参数
         * @param validationParams.rule {Object} 验证规则
         * @param validationParams.rule.ruleName {String} 验证规则类型名称
         * @param [validationParams.rule.pattern] {String} 规则是 `pattern` 时,  本属性是正则字符串
         * @param [validationParams.rule.property] {String} 验证的字段名称
         * @param validationParams.form {Object} angular form对象
         * @Param validationParams.entityAndGroupName {String} 当前验证规则文件名的名称, 是 `实体名_分组名`
         * @constructor
         */
        function ControlValidator(validationParams) {
            var me = this,
                rule = me.rule = validationParams.rule,
                form = me.form = validationParams.form,
                entityAndGroupName = validationParams.entityAndGroupName,
                ruleName = me.ruleName = rule['ruleName'];

            me.property = rule.property;
            me.properties = rule.properties || [];
            me.matchedControls = me.getMatchedControls();
            me.entityName = entityAndGroupName.split('_')[0];
        }
        ControlValidator.prototype.isNoControlProperty = function (property) {
            return property && !this.form.hasOwnProperty(property) && property.indexOf('.') !== -1;
        };

        ControlValidator.prototype.getMatchedControls = function () {
            var me = this,
                form = me.form,
                property = me.property,
                matchedProperties = angular.copy(me.properties) || [];
            if (me.isNoControlProperty(me.property)) {
                angular.forEach(form, function (v, k) {
                    if (k.match(new RegExp(me.property.split('.').join('(\\\[\\\d+\\\])?\.')))) {
                        matchedProperties.push(k);
                    }
                });
            } else if (angular.isString(property) && !!property) {
                matchedProperties.push(property);
            } else if (angular.isArray(matchedProperties) && matchedProperties.length > 0) {
                if (angular.isString(me.property)) {
                    matchedProperties.push(me.property);
                }
            }
            return Arrays.transform(matchedProperties, function (property) {
                return form[property];
            });
        };

        ControlValidator.prototype.apply = function (predicatedCallback) {
            var me = this,
                form = me.form,
                matchedControls = me.matchedControls,
                ruleName = me.ruleName, control, rulePredicate;

            angular.forEach(matchedControls, function (matchedControl) {
                if (matchedControl) {
                    var willSetValidityControls = me.getWillSetValidityControls(matchedControl);
                    rulePredicate = me[ruleName];
                    matchedControl.$parsers.unshift(function (value) {
                        // 异步验证， 需要传入control
                        if (ruleName === 'unique') {
                            rulePredicate.call(me, value, willSetValidityControls);
                        } else {
                            me.defaultControlValidCallback(rulePredicate.call(me, value), willSetValidityControls);
                        }
                        return value;
                    });
                    matchedControl.$formatters.unshift(function (value) {
                        // 异步验证， 需要传入control
                        if (ruleName === 'unique') {
                            rulePredicate.call(me, value, willSetValidityControls);
                        } else {
                            var isValid = rulePredicate.call(me, value);
                            me.defaultControlValidCallback(isValid, willSetValidityControls);
                        }
                        return value;
                    });
                }
            });
        };

        ControlValidator.prototype.getWillSetValidityControls = function (matchedControl) {
            var me = this,
                form = me.form,
                property = me.property,
                properties = me.properties;
            if (!!property && form.hasOwnProperty(property)) {
                return [form[property]];
            } else if (!!property && !form.hasOwnProperty(property) && !!matchedControl) {
                return [matchedControl];
            } else {
                return Arrays.transform(properties, function (property) {
                    return form[property];
                });
            }
        };

        ControlValidator.prototype.defaultControlValidCallback = function (isValid, controls) {
            var ruleName = this.ruleName;
            angular.forEach(controls, function (control) {
                if (control) {
                    control.$setValidity(ruleName, isValid);
                }
            });
        };

        ControlValidator.prototype.notNull = function (rawValue) {
            return rawValue !== null;
        };

        ControlValidator.prototype.notEmpty = function (rawValue) {
            return angular.isString(rawValue) && !!rawValue;
        };

        ControlValidator.prototype.notBlank = function (rawValue) {
            return angular.isString(rawValue) && $.trim(rawValue).length > 0;
        };

        ControlValidator.prototype.pattern = function (rawValue) {
            return !rawValue || (new RegExp(this.rule.pattern).test(rawValue));
        };

        ControlValidator.prototype.min = function (rawValue) {
            var value = Number(rawValue);
            return !rawValue || (angular.isNumber(value) && value >= this.rule.min);
        };

        ControlValidator.prototype.max = function (rawValue) {
            var value = Number(rawValue);
            return !rawValue || (angular.isNumber(value) && value <= this.rule.max);
        };

        ControlValidator.prototype.length = function (rawValue) {
            return !rawValue || (angular.isString(rawValue) && rawValue.length <= this.rule.max && rawValue.length >= this.rule.min);
        };

        ControlValidator.prototype.numeric = function (rawValue) {
            var value = Number(rawValue),
                token = rawValue.split('.');
            if (rawValue) {
                if (this.rule.precision >= 0 && (token[1] || '').length > this.rule.precision) {
                    return false;
                } else if (this.rule.sign === 'POSITIVE' && value < 0) {
                    return false;
                } else if (this.rule.sign === 'NEGATIVE' && value > 0) {
                    return false;
                }
            }
            return true;
        };

        ControlValidator.prototype.range = function (rawValue) {
            var value = Number(rawValue);
            return !rawValue || (angular.isNumber(value) && value >= this.rule.min && value <= this.rule.max);
        };

        ControlValidator.prototype.allMatch = function (value) {
            var me = this,
                form = me.form,
                properties = Arrays.filter(me.properties, function (property) {
                    return !!form[property];
                }),
                i, len, property;
            for (i = 0, len = properties.length; i < len; i++) {
                property = properties[i];
                if (value !== form[property].$viewValue) {
                    return false;
                }
            }
            return true;
        };

        ControlValidator.prototype.compare = function (rawValue) {
            var rule = this.rule,
                than = rule.than,
                compareToVal = this.form[rule.compareTo].$viewValue;
            if (rawValue && compareToVal) {
                if (than === 'GREATER' && rawValue < compareToVal) {
                    return false;
                } else if (than === 'LESS' && rawValue > compareToVal) {
                    return false;
                }
            }
            return true;
        };

        ControlValidator.prototype.unique = function (value, controls) {
            var me = this,
                entityName = me.entityName,
                property = me.property;
            if (value) {
                if (me.reValidTimer) {
                    $timeout.cancel(me.reValidTimer);
                    me.reValidTimer = undefined;
                }
                me.reValidTimer = $timeout(function () {
                    $http.get((window.basePath || '') + '/validation/exists/' + entityName + '/' + property + '/' + encodeURIComponent(value))
                        .success(function (result) {
                            me.defaultControlValidCallback(result && result.success === true, controls);
                            me.reValidTimer = undefined;
                        });
                }, 300);
            } else {
                me.defaultControlValidCallback(true, controls)
            }
        };

        ControlValidator.prototype.eitherNotBlank = function (value) {
            var me = this,
                properties = me.properties,
                form = me.form,
                existsControlProperties = Arrays.filter(properties, function (property) {
                    return !!form[property];
                }),
                existsControls = Arrays.transform(existsControlProperties, function (property) {
                    return form[property];
                }),
                values = Arrays.transform(existsControls, function (control) {
                    return control.$viewValue;
                }),
                notBlankValue = Arrays.findOne(values, function (value) {
                    return angular.isString(value) && $.trim(value).length > 0;
                });
            return notBlankValue !== undefined;
        };

        return {
            newInstance: function (validationParams) {
                return new ControlValidator(validationParams);
            },
            isSupported: function (validationParams) {
                var rule = validationParams.rule;
                return rule && angular.isFunction(ControlValidator.prototype[rule.ruleName]) && validationParams.entityAndGroupName;
            }
        };
    };
})
;